#!/usr/bin/env python
'''
Command line entry point for ESG Replication Client data publication requests.
See the L{main} method for a description of the command line arguments.

Created on May 27, 2010

@author: cward
@organization: USC Information Sciences Institute
'''
import sys
import optparse
import logging

from esgcet.model import PUBLISH_DATASET_EVENT
from esgcet.exceptions import ESGPublishError
from esgcet.config import loadConfig

import replication.configurationKeys as ckeys
from replication.tdsCatalog import get_source_catalog_for_dataset

LOG_FILENAME = 'replicapublication.log'
'@type: string'
logging.basicConfig(filename=LOG_FILENAME,
                    format='%(asctime)s %(levelname)-10s %(funcName)s: %(message)s',
                    filemode='w',
                    level=logging.ERROR)
_top_log   = logging.getLogger('app')
"""
@type: logging.Logger
"""
_top_log.setLevel(logging.ERROR)
_trace_log = logging.getLogger('app.' + __name__)
"""
@type: logging.Logger
"""
verbose = True
"""
Flag to control output to standard out.
@type: boolean
"""
#------------------------------------------------------------------------------
# Functions to validate input arguments.
#------------------------------------------------------------------------------

def _sane_scan(params, config):
    """
    Check that the necessary parameters are present for a scanDirectory
    operation.
    @param params: Parameters from the command line.
    @type params: optparse.OptionParser
    @param config: Configuration data
    @type config: ConfigParser
    @rtype: boolean
    """
    if verbose: print "Checking arguments for a dataset scan ...",
    if (params.target is not None and params.dataset is not None and
        params.map is not None and params.project is not None):
        if verbose: print "pass."
        return True
    if verbose: print "fail!"
    return False

def _sane_generate(params, config):
    """
    Check that the necessary parameters for generating a new catalog for the 
    replicated data set are present.
    @param params: Parameters from the command line.
    @type params: optparse.OptionParser
    @param config: Configuration data
    @type config: ConfigParser
    @rtype: boolean
    """
    if verbose: print "Checking arguments for generating a catalog ...",
    if (params.map is not None and params.source is not None and 
        params.dataset is not None and params.project is not None and
        params.version is not None):
        if verbose: print "pass."
        return True
    if verbose: print "fail!"
    return False

def _sane_publish(params, config):
    """
    Check that the necessary parameters for publishing a replicated dataset
    are present.
    @param params: Parameters from the command line.
    @type params: optparse.OptionParser
    @param config: Configuration data
    @type config: ConfigParser
    @rtype: boolean
    """
    if verbose: print "Checking arguments for publishing replica...",
    if (params.replica is not None and params.dataset is not None and
        params.version is not None and params.project is not None):
        if verbose: print "pass."
        return True
    if verbose: print "fail!"
    return False

def _sane_compare(params, config):
    """
    Check that the necessary parameters for comparing two catalogs are present.
    @param params: Parameters from the command line.
    @type params: optparse.OptionParser
    @param config: Configuration data
    @type config: ConfigParser
    @rtype: boolean
    """
    if verbose: print "Checking arguments for comparing two catalogs ...",
    if params.replica is not None and params.source is not None:
        if verbose: print "pass."
        return True
    if verbose: print "fail!"
    return False

def _sane_check(params, config):
    """
    Check that the necessary parameters to run a pre-publication check are
    present. A pre-publication check runs the equivalent of scan, generate, 
    and compare operations.
    @param params: Parameters from the command line.
    @type params: optparse.OptionParser
    @param config: Configuration data
    @type config: ConfigParser
    @rtype: boolean    
    """
    if verbose: print "Checking arguments for running a check operation ...",
    if (params.target is not None and params.dataset is not None and
        params.project is not None and params.version is not None):
        if verbose: print "pass."
        return True
    if verbose: print "fail!"
    return False

#------------------------------------------------------------------------------
# Functions to call API
#------------------------------------------------------------------------------
def perform_scan(target, project, map, dataset, config):
    """
    Wrapper for Publication Replication API scan function.
    @param target: Path to directory to scan.
    @type target: string
    @note: The called publish scan method accepts a list of directories to scan;
    however, replication currently only supports one.
    @param project: Name of the project with the files.
    @type project: string
    @param map: Path to write the resulting map file.
    @type map: string
    @param dataset: ID for the dataset of the scanned directory
    @type dataset: string
    @param config: Configuration data object.
    @type config: ConfigParser
    @return: Status of the scan attempt.
    @rtype: boolean
    """
    _trace_log.info("Enter")
    status = False
    from esgcet.publish.replica import scanDirectory
    try:
        if verbose: print "Scanning %s for dataset %s" %(target, dataset)
        # scanDirectory wants a list of directories so give it a list of one
        scanDirectory([target], project, map, dataset)
        status = True
        if verbose: print "Map in %s" %(map)

    except ESGPublishError, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("ESGPublishError:")

    except EnvironmentError, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("EnvironmentError:")

    _trace_log.info("Return %s" %(status))
    return status

def generate_replica_catalog(map, dataset, sourceXml, project, config, 
                             service=None, version=None, sourceGateway=None):
    """
    Wrapper for Publication Replication API generate replica function.
    @param map: Path to the file with the data map to use to generate the
    catalog.
    @type map: string
    @param dataset: ID for the dataset.
    @type dataset: string
    @param sourceXml: Path to the originating TDS catalog.
    @param sourceXml: string
    @param project: Name of the project to which the dataset belongs.
    @type project: string
    @param config: Configuration data.
    @type config: ConfigParser
    @param service: String service name, matching a service identifier in the 
    configuration file. Specifies the file lister to be used.
    @param version: Version number to assign this replicated dataset.
    @type version: integer
    @param sourceGateway: Code name for the gateway used by the originating data node.
    @type sourceGateway: string
    @return: Status of the operation and name of generated catalog file.
    @rtype: tuple (boolean, string)
    """
    _trace_log.info("Enter")
    status = False
    rep_cata = None
    
    from esgcet.publish.replica import generateReplicaThreddsCatalog
    from esgcet.publish import CREATE_OP

    if (sourceGateway is None and 
        config.has_option(ckeys.SECTION, ckeys.SOURCE_GTWY_KEY)):
        sourceGateway = config.get(ckeys.SECTION, ckeys.SOURCE_GTWY_KEY)
    resultDictionary = dict()
    sourceCatalog = dict()
    sourceCatalog[dataset] = sourceXml

    try:
        if verbose: print ("Generating catalog for dataset %s using map %s" %
                           (dataset, map))
        resultDictionary = generateReplicaThreddsCatalog(map,
                                                         CREATE_OP,
                                                         sourceCatalog,
                                                         sourceGateway=sourceGateway,
                                                         service=service,
                                                         datasetId=dataset,
                                                         version=version,
                                                         project=project)
        rep_cata = _save_catalog(resultDictionary, config)
        status = True

    except ESGPublishError, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("ESGPublishError:")

    except EnvironmentError, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("EnvironmentError:")

    except Exception, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("Exception:")

    _trace_log.debug(resultDictionary)
    _trace_log.info("Return (%s, %s)" % (status, rep_cata))
    return (status, rep_cata)

def publish_replica(catalog, dataset, project, version, config):
    """
    Wrapper for Publication Replication API to publish dataset to a gateway.
    @param catalog: Path to the replica TCS catalog XML file.
    @type catalog: string
    @param dataset: ID for dataset.
    @type dataset: string
    @param project: Name for the project of this dataset.
    @type project: string
    @param version: Version for the dataset
    @type version: integer
    @param config: Configuration data.
    @type config: ConfigParser
    @return: Status of the operation.
    @rtype: boolean
    """
    _trace_log.info("Enter")
    status = False

    from esgcet.publish.replica import publishCatalogs

    try:
        cat = open(catalog,"r")
        catalogXML = cat.read()
        cat.close()
        catalogDict = {}
        dsetKey = (dataset, version)
        catalogDict[dsetKey] = catalogXML
        parentDict = {}
        parentDict[dataset] = project
        statusDict = publishCatalogs(catalogDict,parentDict,
                                     las=False,publish=True)
        statusFlag = statusDict[dsetKey]
        status = (statusFlag == PUBLISH_DATASET_EVENT)

    except ESGPublishError, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("ESGPublishError:")

    except EnvironmentError, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("EnvironmentError:")

    except Exception, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("Exception:")

    _trace_log.info("Return %s" % (status))
    return status

def compare_catalog_files(original_catalog, replica_catalog, config):
    """
    Compare two catalog files.
    @param original_catalog: Path to original TDS catalog XML file.
    @type original_catalog: string
    @param replica_catalog: Path to generated replica catalog XML file.
    @type replica_catalog: string
    @param config: Data from the configuration file.
    @type config: ConfigParser
    @return: Status of comparison, True if the same under the rules for
    replication.
    @rtype: boolean
    """
    _trace_log.info("Enter: compare %s to %s" % 
                    (original_catalog, replica_catalog))
    from replication.comparisons import createContext, extractDataset, DatasetExtract

    if config.has_option(ckeys.SECTION,ckeys.COMPARE_CHECKED):
        properties = config.get(ckeys.SECTION,ckeys.COMPARE_CHECKED).split(',')
        _trace_log.debug("Setting DatasetExtract.checked_properties to %s" 
                         %(properties))
        newset = set()
        for property in properties:
            newset.add(property.strip()) # add with any extra whitespace removed
        DatasetExtract.set_checked_properties(newset)
    if config.has_option(ckeys.SECTION,ckeys.COMPARE_REQUIRED):
        properties = config.get(ckeys.SECTION,ckeys.COMPARE_REQUIRED).split(',')
        _trace_log.debug("Setting DatasetExtract.checked_properties to %s" 
                         %(properties))
        newset = set()
        for property in properties:
            newset.add(property.strip()) # add with any extra whitespace removed
        DatasetExtract.set_required_properties(newset)

    status = False

    try:
        context = createContext(original_catalog)
        orig_dataset = extractDataset(context,ignore_aggregation=True, id=None)
        context = createContext(replica_catalog)
        replica_dataset = extractDataset(context,ignore_aggregation=True, id=None)
        status = (orig_dataset == replica_dataset)

    except EnvironmentError, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("EnvironmentError:")

    except Exception, e:
        print >>sys.stderr, str(e)
        _trace_log.exception("Exception:")

    _trace_log.info("Return: %s" %(status))
    return status;

def check_dataset(scandir, dataset, project, version, config, 
                  service=None, sourceGateway=None):
    """
    Perform a scan, generate a replica catalog, and compare it to the 
    originating source catalog. When this function is run, the map file gets
    the name "project_name_mapfile.txt" where "project_name" is the string found
    in the "project" parameter. The source catalog file is looked up in the
    text file database created by L{esgreplicate} when it setup the the 
    download. (Also see L{tdsCatalog.mapdb_file}.)
    @param scandir: Path to directory with dataset files to scan.
    @type scandir: string
    @param dataset: ID for the dataset being scanned.
    @type dataset: string
    @param project: Project name for the dataset.
    @type project: string
    @param version: Version number for the dataset files.
    @type version: integer
    @param config: Configuration data.
    @type config: ConfigParser
    @param sourceGateway: Code name for the gateway used by the originating data node.
    @type sourceGateway: string
    @return: Results of the check.
    @rtype: boolean
    """
    _trace_log.info("Enter: %s, %s, %s, %d" %(scandir, dataset, project, version))
    tmpdir = './'
    if config.has_option(ckeys.SECTION, ckeys.TDS_TEMP_DIR_KEY):
        tmpdir = config.get(ckeys.SECTION, ckeys.TDS_TEMP_DIR_KEY)
    if tmpdir.endswith('/'):
        map = tmpdir + project + "_mapfile.txt"
    else:
        map = tmpdir + '/' + project + "_mapfile.txt"
    
    try:
        status = perform_scan(scandir, project, map, dataset, config)
        if status:
            src_cat = get_source_catalog_for_dataset(dataset, tmpdir)
            status, replica = generate_replica_catalog(map,
                                                   dataset,
                                                   src_cat,
                                                   project,
                                                   config,
                                                   service,
                                                   version,
                                                   sourceGateway)
            if status:
                status = compare_catalog_files(src_cat,replica,config)
                if status == False:
                    print >>sys.stderr,("Catalogs %s and %s do not match." 
                                        %(src_cat, replica))
                else:
                    print >>sys.stderr, "Generation of replica catalog failed."
        else:
            print >>sys.stderr, "Scan of %s failed" %(scandir)
    except ESGPublishError, e:
        _trace_log.exception("ESGPublishError:")
        print >>sys.stderr, "ESGPublishError: ", str(e)
        status = False
    except EnvironmentError, e:
        _trace_log.exception("EnvironmentError:")
        print >>sys.stderr, str(e)
        status = False
    
    _trace_log.info("Return: %s" %(status))
    return status

def _save_catalog(catalogDic,config):
    """
    Save a generated catalog to a file.
    @param catalogDic: Mapping dataset and version to catalog XML.
    @type catalogDic: dictionary
    @param config: Configuration data
    @type config: ConfigParser
    @return: Returns the name of the file saved.
    @rtype: string
    @raise IOError: May raise an IO error when writing file to disk.    
    """
    saveDir = config.get(ckeys.SECTION, ckeys.TDS_TEMP_DIR_KEY)
    datasets = catalogDic.keys()
    for dataset in datasets:
        name, version = dataset
        outname = saveDir + "/" + name + ".replica.v" + str(version) + ".xml"
        out = open(outname,"w")
        out.write(catalogDic[dataset])
        out.close
    return outname
    
def main(argv):
    """
    Prepare a downloaded dataset for publication to the gateway as a replica.
    
    At least two steps are required for publication of a replica after the data
    has been downloaded. The first step performs a scan of the files, generates
    a replicated TDS catalog for the dataset and verifies that it matches the
    originating TDS catalog. These steps may also be performed separately. The
    second step publishes the data to the gateway as as replicated dataset.
    
    @param argv: Command line arguments. Options are::
        -a, --action ACTION:
               Action to perform for this execution (required). Available
               actions are scan, catalog, compare, check, and publish.
        -c, --config CONFIG_FILE:
               Path to a configuration file; overrides default ESG INI file.
        -d, --dataset DATASET_ID:
               A dataset id to act upon.
        -g, --gateway ESG_CODE:
               Code name of the gateway of the originating data node.
        -l, --service service name:
               String service name, matching a service identifier in the 
               configuration file. Specifies the file lister to be used.
        -m, --map MAP_FILE:
               a map file to either read or write.
        -p, --project PROJECT:
               Name of the project for the dataset id.
        -q, --quiet:
               Do not print status messages to the screen.
        -r, --replica CATALOG_FILE:
               Path to the catalog file generated or to generate for the replica,  
        -s, --source CATALOG_FILE:
               Path to the catalog file downloaded from the source data node.
        -t, --target SCAN_DIRECTORY:
               Directory with dataset files to use for generating a catalog or
               publication.
        -v, --version NUMBER:
               Version number to assign the replicated dataset.
        -h, --help:
               Print some help text (via optparse).
    Some arguments will take on different meanings depending of the context of
    the action to be performed.
    
    @return: Status of operation, 0 for success, non-zero for an error.
    @rtype: integer
    """

    status = False
    
    p = optparse.OptionParser()
    p.add_option("-a", "--action", action="store", type="choice", dest="action",
                 choices=["scan", "catalog", "compare", "check", "publish"],
                 help="Action or step to perform for replica publication: \
allowed actions are scan, catalog, compare, check, and publish")
    p.add_option("--config", "-c", action="store", type="string",
                 help="Location of configuration file.")
    p.add_option( "-d", "--dataset", type="string",
                  help="Dataset id to replicate.")
    p.add_option("-g", "--gateway", type="string",
                 help="Code name for the gateway of the originating data node.")
    p.add_option("-l", "--service", type="string",
                 help="String service name, matching a service identifier in the configuration file. Specifies the file lister to be used.")
    p.add_option("-m", "--map", 
                 help="Scanned map file, for input or output")
    p.add_option("-p", "--project", help="Project name.")
    p.add_option("-r", "--replica", 
                 help="Catalog file for replica, input or output.")
    p.add_option("-s", "--source", help="Catalog from the source data node.")
    p.add_option("-t", "--target", help="Target directory for scan")
    p.add_option("-v", "--version",
                 help="Version to assign published replica dataset")
    p.add_option("-q", "--quiet", action="store_true",
                 help="Do not print status messages to the screen.")

    (opts, args) = p.parse_args(argv) # args ignored
    
    if len(args) != 0:
        print >>sys.stderr, "Warning: ignored arguments %s" %(args)

    config = loadConfig(opts.config) # Load ESG INI as default

    runtime_log_level = config.get(ckeys.SECTION, ckeys.LOG_LEVEL, default=None)
    if runtime_log_level is not None:
        _top_log.setLevel(logging.getLevelName(runtime_log_level))

    _trace_log.info("Command line and configuration processed.")
    _trace_log.debug("command line='%s'", argv)
    
    if opts.quiet is not None:
        global verbose;
        verbose = False

    if opts.action is None:
        print >>sys.stderr, "Error: An action must be specified."
    elif opts.action == 'scan':
        if _sane_scan(opts,config):
            status = perform_scan(opts.target, opts.project, opts.map, 
                                  opts.dataset,config)
        else:
            print >>sys.stderr, "A scan needs target to scan, a dataset, a \
project name, and an output map file."
    elif opts.action == 'catalog':
        if _sane_generate(opts,config):
            status, replica = generate_replica_catalog(opts.map,
                                                       opts.dataset,
                                                       opts.source,
                                                       opts.project,
                                                       config,
                                                       opts.service,
                                                       int(opts.version),
                                                       opts.gateway)
            if status and verbose:
                print "Replica catalog stored in %s" %(replica)
 
        else:
            print >>sys.stderr, "Need source catalog, a dataset, a project, \
and a map file for input."
    elif opts.action == 'publish':
       if _sane_publish(opts,config):
            status = publish_replica(opts.replica,
                                     opts.dataset,
                                     opts.project,
                                     int(opts.version),
                                     config)
            if status is True and verbose:
                print ("Replica of %s published successfully for project %s" %
                       (opts.dataset, opts.project))
       else:
           print>>sys.stderr,("Need a replica catalog, dataset, project and version to publish.")

    elif opts.action == 'compare':
        if _sane_compare(opts, config):
            status = compare_catalog_files(opts.source, opts.replica, config)
            if status and verbose:
                    print ("%s and %s match." % (opts.source, opts.replica))
            else:
                if verbose: 
                    print ("%s and %s do not match." % (opts.source, opts.replica))
        else:
            print >>sys.stderr, ("Need source and replica catalogs to compare.")
    elif opts.action == 'check':
        if _sane_check(opts, config):
            status = check_dataset(opts.target,
                                   opts.dataset,
                                   opts.project,
                                   int(opts.version),
                                   config,
                                   opts.gateway)
            if status and verbose:
                print ("Dataset %s passes \"check\" tests." % 
                       (opts.dataset))
            else:
                if verbose: 
                    print("Dataset %s fails \"check\" tests." %
                          (opts.dataset))
        else:
            print >>sys.stderr, "Need target directory, dataset, project, and version."
    else: # never happen in production; help in development
        raise NotImplementedError("Error: action value '" + 
                                  opts.action + "' not implemented.")

    _trace_log.info("Exit: %s" %(status))
    if status is True:
        if verbose: print "Success"
        return 0
    if verbose: print "Failure"
    return 1

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))