#!/usr/bin/env python
'''
Command line entry point for the ESG Replication Client data movement requests.

See the L{main} module for command line options.

@author: ebrady, cward
@organization: USC Information Sciences Institute
'''
import sys
import optparse
import logging

from esgcet.config import loadConfig

import replication.configurationKeys as ckeys
from replication.dataMovementRequest import create_dmr_file
from replication.fileAccess import get_dataset_listing

LOG_FILENAME = 'replication.log'
'@type: string'
# Possible alternative format for trace log records
#                    format='%(levelname)-10s %(funcName)s: %(message)s',
logging.basicConfig(filename=LOG_FILENAME,
                    format='%(asctime)s %(levelname)-10s %(funcName)s: %(message)s',
                    filemode='w',
                    level=logging.ERROR)
_top_log = logging.getLogger('app')
"""
@type: logging.Logger
"""
_top_log.setLevel(logging.ERROR)
_trace_log = logging.getLogger('app.' + __name__)
"""
@type: logging.Logger
"""
__verbose = True
"""
Flag to control output to standard out.
@type: boolean
"""

def esgreplicate(parameters, config):
    """
    Produces a data transfer request based on the parameters given.
    
    @param parameters: Parsed arguments to the program.
    @type parameters: dictionary.
    @param config: Configuration dictionary.
    @type config: ConfigParser object.
    @return: Status of operation.
    @rtype: boolean
    """
  
    if __verbose: print "Retrieving meta data from gateway..."
    # Get list of dataset ids to use for query to gateway
    list_of_dataset_ids = get_dataset_listing(parameters["input_file"], config)
  
    if __verbose: print "Processing meta data to generate download request..."
    # Create the XML output file for the data movement client
    successful = create_dmr_file(parameters, config, list_of_dataset_ids)
  
    if not successful:
      if __verbose: 
          print "Output request file could not be successfully created."
    else:
      if __verbose: 
          print "Request successfully created, located at " + parameters["output_file"]
  
    return successful

#------------------------------------------------------------------------------
#	main
#------------------------------------------------------------------------------
def main(argv):
    """
    Parse the command line arguments to create a dictionary object for use by
    called subroutines. The two required arguments are for input and output file
    names.
    
    Using I{dataset ids} specified in an input file, this program creates a
    request for use by a separate I{data movement agent}. Data sets are input
    with a text file with dataset ids (models, runs, files, etc), one per line.

    @param argv: A sequence of strings containing options and values to assign to
    each.
    Options are::
      -c, --config NAME: 
                 Name of configuration file; overrides the default ESG INI file.
      -i, --input NAME:
                 Provide address of input file containing the dataset ids to be
                 mirrored.  Each line of the input file is a single dataset
                 identifier.
      -t --type NAME:
                 Download_type, e.g. bdm, globus-url-copy, globus-online, list
                 (guc and gol are short forms for globus-url-copy and globus 
                 online respectively); default is bdm. (Can 
                 also be set in the configuration file.) The dml option 
                 will not be rejected, but this agent is not supported;
                 use at your own risk.
      -o, --output NAME: 
                 Provide address of output file where data transfer request will 
                 be written.
      -q, --quiet:
                 Do not print status messages to the screen.
      -r, --replica_root:
                 Target root directory for downloaded datasets to be replicated.
      -h, --help:
                 Print some help text (via optparse).
    
    Note that initially the only supported URL-type for a data movement agent 
    is gsiftp.
    @return: Status of operation with 0 indicating success. 
    @rtype: integer
    """

    p = optparse.OptionParser(usage="%prog -i file -o file [-c cfg_file]" +
                              " [-t type] [-r path] [-q]")
    
    p.add_option("-c", "--config", action="store",
                 dest="config_file",
                 help="Path to a configuration file. Default is the ESG INI file.")
    p.add_option("-i", "--input", action="store", dest="input_file",
                 help="Path to file with list of datasets to process.")
    p.add_option("-t", "--type", dest="download_type",
                 action="store", default='bdm',
                 type="choice", choices=['bdm', 'guc', 'globus-url-copy', 'gol',
                                          'globus-online', 'list', 'dml'],
                 help="Target data movement agent. Selections include" +
                      " 'globus-url-copy,' 'globus-online', 'list,' and 'dml.'" +
                      " Default is 'bdm.' See documentation for complete list.")
    p.add_option("-o", "--output", action="store", dest="output_file",
                 help="Path to write the generated request file.")
    p.add_option("-r", "--replica_root", action="store", dest="replica_root",
                 help="Path on disk to root directory for replicated data.",
                 default=None)
    p.add_option("-q", "--quiet", action="store_true",
                 help="Do not print status messages to the screen.")

    (options, lastargs) = p.parse_args()
  
    if options.input_file is None:
      print >>sys.stderr, "An input file of dataset ids is required."
      return 2
  
    if options.output_file is None:
      print >>sys.stderr, "An output file for the data movement request is required."
      return 2
  
    if options.quiet is not None:
        global __verbose
        __verbose = False

  #===============================================================================
  # An earlier revision did not use optparse and created a dictionary from the
  # command line arguments. This re-creates that dictionary object that code
  # further down the calling sequence expects rather than changing that code too.
  #===============================================================================
    parameters = dict()
    parameters['input_file'] = options.input_file
    parameters['output_file'] = options.output_file
    parameters['config_file'] = options.config_file
    parameters['download_type'] = options.download_type
    parameters['replica_root'] = options.replica_root
    parameters['quiet'] = __verbose
  
    _trace_log.debug("input_file=%(input_file)s output_file=%(output_file)s download_type=%(download_type)s config_file=%(config_file)s", parameters)
    _trace_log.debug("lastargs = %s", lastargs)

    config = loadConfig(options.config_file)
  
    runtime_log_level = config.get(ckeys.SECTION, ckeys.LOG_LEVEL, default=None)
    if runtime_log_level is not None:
        _top_log.setLevel(logging.getLevelName(runtime_log_level))

    _trace_log.info("Command line and configuration processed.")
 
    status = esgreplicate(parameters, config)
  
    _trace_log.info("Return: status = %s" %(status))
    if status is True:
        if __verbose: print "Success"
        return 0
    if __verbose: print "Failure"
    return 1

if __name__ == "__main__":
  sys.exit(main(sys.argv[1:]))
