"""
File I/O routines. 
@author: ebrady, cward
@organization: USC Information Sciences Institute
"""
import os
import sys
import logging
from types import StringType

_trace_log = logging.getLogger('app.' + __name__)
"""
@type: logging.Logger
"""
#------------------------------------------------------------------------------
#	__open_file
#	Accepts:  	string filename, string access_type
#			access_type is one of "a", "r", or "o" (for "append", 
#			"read", or "overwrite")
#	Returns:	File object
#------------------------------------------------------------------------------
def __open_file(filename, access_type):
  """
  Open a file after checking the arguments and access rights.
  @param filename: Path to the file.
  @type filename: string
  @param access_type: Requested access rights.
  @type access_type: string
  @return: File object
  @rtype: file
  @raise RuntimeError: Access check and file opening may throw IO exceptions.
  @deprecated: Trap the open call directly without this extra stuff
  """
  _trace_log.info("Enter: %s, %s", filename, access_type)
  file_ref = None
  if check_file_access(filename, access_type):
    file_ref = open(filename, access_type)
  else:
      raise RuntimeError("Invalid access for %s" %(filename))
  _trace_log.info("Exit")
  return file_ref

#------------------------------------------------------------------------------
#	check_file_access
#	Accepts:  	string filename, string access_type
#			access_type is one of "a", "r", or "o" (for "append", 
#			"read", or "overwrite")
#	Returns:	Boolean, where "True" means the filename is valid and 
#			can be accessed as specified
#------------------------------------------------------------------------------
def check_file_access(filename, access_type):
  """
  Check that the arguments are strings and check the access rights against the
  requested access.
  @param filename: Path to file
  @type filename: string
  @param access_type: Access that will be requested.
  @type access_type: string
  @return: True if the access will be granted, else False
  @rtype: boolean
  @deprecated: Trap the open call directly without this extra stuff
  """
  if type(filename) is not StringType:
    print >>sys.stderr, "Filename provided is not a string."    
    return False

  if type(access_type) is not StringType:
    print >>sys.stderr, "Access type provided is not a string (must be either \"a\", \"r\", or \"o\""
    return False

  access = False
  if access_type is "a" or access_type is "o":
    if os.path.exists(filename):
      return os.access(filename, os.W_OK)
    return True
  if access_type is "r":
    if not os.access(filename, os.F_OK):
      print >>sys.stderr,"Filename provided does not exist, or is not a file"
      return False
    return os.access(filename, os.R_OK)

  return access

#------------------------------------------------------------------------------
#	get_dataset_listing
#	Accepts:	string input_file
#	Returns:	List of dataset ids
#------------------------------------------------------------------------------
def get_dataset_listing(input_file, config):
  """
  Read the dataset IDs from the input file into a list object.
  @param input_file: Path to file with dataset IDs.
  @type input_file: string
  @param config: Configuration data.
  @type config: ConfigParser
  @return: The dataset IDs as a list.
  @rtype: list
  """
  _trace_log.info("Entry: %s", input_file)
  list_of_dataset_ids = []
  try:
      dataset_file = open(input_file, "r")
      for line in dataset_file:
          if line[0] == '#':
              _trace_log.debug("Skipping line: %s", line)
              continue
          _trace_log.debug("Processing line: %s", line)
          datasetId = line.strip()
          if datasetId is not None:
             list_of_dataset_ids.append(datasetId);
      dataset_file.close()
  except EnvironmentError, e:
      _trace_log.exception("Error reading file of dataset IDs:")
      print >>sys.stderr, "Error reading file of dataset IDs:", str(e)
      
  _trace_log.debug("list_of_dataset_ids: %s", list_of_dataset_ids)

  if not list_of_dataset_ids:
    print >>sys.stderr, "No dataset IDs found in input file."

  _trace_log.info("Return: %s", list_of_dataset_ids)
  return list_of_dataset_ids
