'''
Routines for parsing XML data from the ESG meta data servers.
@author: ebrady, cward
@organization: USC Information Sciences Institute
@todo: A possible enhancement or optimization could be accepting the XML
documents as Python XML objects instead of string objects.
'''
import logging
from sys import stderr
from lxml import etree

_trace_log = logging.getLogger('app.' + __name__)
"""
@type: logging.Logger
"""
def getLocationList(xml_string, protocol="GridFTP"):
    """
    Parse the ESG meta data to extract the files in the data set. The returned
    list includes the URL protocol. The meta data may have multiple access
    protocols defined for a file. Use the value passed as protocol as selector.
    @param xml_string: The XML meta data.
    @type xml_string: string
    @param protocol: Server protocol to select for. Default is "GridFTP."
    @type protocol: string
    @return: A list of the URLs for the files in a data set.
    @rtype: list 
    """
    _trace_log.info("Entry: Seeking %s data_access_capability" % (protocol))
    list_of_files = []
    metadata_query_result = etree.fromstring(xml_string)
    esg_namespace = metadata_query_result.nsmap.get("esg")
    base_uri_element = metadata_query_result.xpath('/t:ESG/t:dataset/t:data_access_capability', namespaces={'t': esg_namespace})
    base_uri = None
    for element in base_uri_element:
        protocol_type = element.get('type')
        _trace_log.debug("Comparing %s to %s" % (protocol_type, protocol))
        if protocol_type.lower() != protocol.lower():
            continue
        base_uri = element.get('base_uri')
        capability_name = element.get('name')
        break

    if base_uri is None:
        print >> stderr, "No data_access_capability of protocol %s" % (protocol)
        _trace_log.error("Error Return: No data_access_capability of protocol %s" % (protocol))
        return list_of_files

    if base_uri.endswith('/'):
        pass
    else:
        base_uri = base_uri + '/'

    files_to_transfer = metadata_query_result.xpath('/t:ESG/t:dataset/t:files/t:file/t:file_access_point', namespaces={'t': esg_namespace})
    for element in files_to_transfer:
        capability = element.get('data_access_capability')
        name = element.get("uri")
        if name.startswith('/'):
            name = name.lstrip('/')
        _trace_log.debug("Checking capability %s" % (capability_name))
        if capability == capability_name:
            _trace_log.debug("Adding %s to list" % (name))
            list_of_files.append(base_uri + name)

    _trace_log.info("Normal Return")
    return list_of_files

def getSizeList(xml_string):
    """
    Extract file sizes from TDS XML meta data. The XML is passed in as a string.
    An XML document object is created internally. Uses XPath queries.
    @param xml_string: The XML meta data.
    @type xml_string: string
    @return: List of sizes for files.
    @rtype: list
    """
    list_of_sizes = []
    metadata_query_result = etree.fromstring(xml_string)
    esg_namespace = metadata_query_result.nsmap.get("esg")

    sizes_of_files_to_transfer = metadata_query_result.xpath('/t:ESG/t:dataset/t:files/t:file', namespaces={'t': esg_namespace})
    for element in sizes_of_files_to_transfer:
        size = element.get("size")
        list_of_sizes.append(size)

    return list_of_sizes

def getChildrenList(xml_string):
    """
    Returns a list of c in the XML object.
    @param xml_string: XML to query.
    @type xml_string: string
    @return: List of children from the document.
    @rtype: list
    """
    list_of_children = []
    metadata_query_result = etree.fromstring(xml_string)
    esg_namespace = metadata_query_result.nsmap.get("esg")

    children = metadata_query_result.xpath('/t:ESG/t:dataset/t:dataset', namespaces={'t': esg_namespace})
    for child in children:
        name = child.get("id")
        list_of_children.append(name)

    return list_of_children

def getThreddsCatalogUri(xml_string):
    """
    Retrieve the URL of the source data node TDS catalog from the results of a
    meta data query to an ESG gateway node.
    
    @param xml_string: The XML meta data document.
    @type xml_string: string
    @return: The URL to the source data node TDS catalog as a string.
    @rtype: string
    """
    metadata_query_result = etree.fromstring(xml_string)
    esg_namespace = metadata_query_result.nsmap.get("esg")
    dataset = metadata_query_result.xpath("/t:ESG/t:dataset", namespaces={'t': esg_namespace})
    #attributes = dataset.attrib
    #source_uri = attributes["source_catalog_uri"]
    source_uri = dataset[0].get("source_catalog_uri")
    return source_uri

def getMetadataList(xml_string):
    """
    Convert the metadata into a Python list of dictionary objects. The XML is
    passed in as a string. An XML document object is created internally. Uses
    XPath queries. The returned data structure has the format
    C{[file_id, size, checksum, checksum_type, [access_points]]}. Other metadata
    ignored.
    @param xml_string: The XML meta data.
    @type xml_string: string
    @return: Metadata stored as a list of Dictionary objects.
    @rtype: list
    @author: Estanislao Gonzalez requested and supplied the initial code for
    this function.
    """
    list_of_dictionaries = []
    metadata_query_result = etree.fromstring(xml_string)
    esg_namespace = metadata_query_result.nsmap.get("esg")

    sizes_of_files_to_transfer = metadata_query_result.xpath('/t:ESG/t:dataset/t:files/t:file', namespaces={'t': esg_namespace})
    for element in sizes_of_files_to_transfer:
        id = element.get("name")
        size = element.get("size")
        #fap = file access point
        faps = []
        for fap in element:
            faps.append({"uri": fap.get("uri"),"type": fap.get("data_access_capability")})            
        
        checksum = element.get("checksum")
        checksum_type = element.get("checksum_type")
        list_of_dictionaries.append({ "id":id, "size":size, "checksum":checksum, "checksum_type":checksum_type, "access_points":faps})
    
    return list_of_dictionaries
